import unittest
from typing import *

from tomlhold.core import TOMLHolder


class TestTOMLHolder(unittest.TestCase):

    def setUp(self: Self) -> None:
        "This method is called before each test case to set up a new instance of the TOMLHolder class."
        self.initial_data = {
            "key1": "value1",
            "key2": "value2",
            "nested": {"nested_key1": "nested_value1"},
        }
        self.holder = TOMLHolder()
        self.holder.data = self.initial_data

    def test_get_item(self: Self) -> None:
        "This testmethod tests if we can get an item by key."
        self.assertEqual(self.holder["key1"], "value1")
        self.assertEqual(self.holder["nested"]["nested_key1"], "nested_value1")

    def test_set_item(self: Self) -> None:
        "This testmethod tests if we can set an item by key."
        self.holder["key3"] = "value3"
        self.assertEqual(self.holder["key3"], "value3")

    def test_delete_item(self: Self) -> None:
        "This testmethod tests if we can delete an item by key."
        del self.holder["key1"]
        self.assertNotIn("key1", self.holder)

    def test_key_in_holder(self: Self) -> None:
        "This testmethod tests if we can check if a key is in the TOMLHolder."
        self.assertIn("key1", self.holder)
        self.assertNotIn("key3", self.holder)

    def test_len(self: Self) -> None:
        "This testmethod tests if len function returns the correct length."
        self.assertEqual(len(self.holder), len(self.initial_data))
        self.holder["key3"] = "value3"
        self.assertEqual(len(self.holder), len(self.initial_data) + 1)

    def test_update(self: Self) -> None:
        "This testmethod tests if the update method works correctly."
        new_data: dict = {"key3": "value3", "key4": "value4"}
        self.holder.update(new_data)
        self.assertEqual(self.holder["key3"], "value3")
        self.assertEqual(self.holder["key4"], "value4")

    def test_iteration(self: Self) -> None:
        "This testmethod tests if iteration over keys works."
        key: Any
        keys: list = list(self.holder)
        self.assertListEqual(keys, list(self.initial_data.keys()))

    def test_copy(self: Self) -> None:
        "This testmethod tests if copy works correctly."
        holder_copy: TOMLHolder = self.holder.copy()
        self.assertEqual(holder_copy, self.holder)

    def test_equality(self: Self) -> None:
        "This testmethod tests equality of two TOMLHolder instances."
        holder2: TOMLHolder = TOMLHolder(self.initial_data)
        self.assertEqual(self.holder, holder2)
        holder2["key3"] = "value3"
        self.assertNotEqual(self.holder, holder2)

    def test_clear(self: Self) -> None:
        "This testmethod tests if clear method works."
        self.holder.clear()
        self.assertEqual(len(self.holder), 0)

    def test_initialization_with_toml(self: Self) -> None:
        "This testmethod tests if the TOMLHolder initializes correctly with TOML data."
        toml_data: str = """
        [section]
        key1 = "value1"
        key2 = "value2"
        [section.nested]
        nested_key1 = "nested_value1"
        """
        holder: TOMLHolder = TOMLHolder.loads(toml_data)
        self.assertEqual(holder["section", "key1"], "value1")
        self.assertEqual(holder["section"]["nested"]["nested_key1"], "nested_value1")

    def test_invalid_key_access(self: Self) -> None:
        "This testmethod tests if accessing a non-existent key raises KeyError."
        with self.assertRaises(KeyError):
            _: Any = self.holder["non_existent_key"]

    def test_pop(self: Self) -> None:
        "This testmethod tests if pop works correctly."
        value: Any = self.holder.pop("key1")
        self.assertEqual(value, "value1")
        self.assertNotIn("key1", self.holder)

    def test_popitem(self: Self) -> None:
        "This testmethod tests if popitem works correctly."
        key: Any
        value: Any
        key, value = self.holder.popitem()
        self.assertIn(key, self.initial_data)
        self.assertEqual(value, self.initial_data[key])

    def test_keys_values_items(self: Self) -> None:
        "This testmethod tests keys(), values(), and items() methods."
        self.assertListEqual(list(self.holder.keys()), list(self.initial_data.keys()))
        self.assertListEqual(
            list(self.holder.values()), list(self.initial_data.values())
        )
        self.assertListEqual(list(self.holder.items()), list(self.initial_data.items()))


if __name__ == "__main__":
    unittest.main()



import unittest

__all__ = ["test"]


def test() -> unittest.TextTestResult:
    loader: unittest.TestLoader = unittest.TestLoader()
    tests: unittest.TestSuite = loader.discover(start_dir="tomlhold.tests")
    runner: unittest.TextTestRunner = unittest.TextTestRunner()
    result: unittest.TextTestResult = runner.run(tests)
    return result



import tomllib
from datetime import date, datetime, time
from functools import partial
from typing import *

import datahold
import setdoc
import tomli_w
from frozendict import frozendict

__all__ = ["Holder", "TOMLHolder"]


def getdict(d: dict, /, *, freeze: bool = False) -> dict | frozendict:
    "This function returns a TOML dict."
    ans: dict
    k: Any
    msg: str
    ans = dict()
    for k in d.keys():
        if type(k) is not str:
            msg = "type %r is not allowed for keys of dictionaries"
            msg %= type(k).__name__
            raise TypeError(msg)
        ans[k] = getvalue(d[k], freeze=freeze)
    if freeze:
        return frozendict(ans)
    else:
        return ans


def getkey(key: int | str) -> int | str:
    "This function returns a TOML key."
    msg: str
    if type(key) is int:
        return key
    if type(key) is str:
        return key
    msg = "type %r is not allowed for keys"
    msg %= type(key).__name__
    raise TypeError(msg)


def getkeys(keys: Any, /) -> list[int | str]:
    "This function returns TOML keys."
    if isinstance(keys, tuple):
        return list(map(getkey, keys))
    else:
        return [getkey(keys)]


def getvalue(value: Any, /, *, freeze: bool = False) -> Any:
    "This function returns a TOML value."
    msg: str
    g: Iterable
    t: str
    if isinstance(value, (dict, frozendict)):
        return getdict(value, freeze=freeze)
    if isinstance(value, (list, tuple)):
        g = map(partial(getvalue, freeze=freeze), value)
        if freeze:
            return tuple(g)
        else:
            return list(g)
    for t in (bool, float, int, str):
        if isinstance(value, t):
            return t(value)
    for t in (datetime, date, time):
        if type(value) is t:
            return value
    msg = "type %r is not allowed for values"
    msg %= type(value).__name__
    raise TypeError(msg)


class TOMLHolder(datahold.OkayDict):
    @setdoc.basic
    def __delitem__(self: Self, keys: tuple | int | str) -> None:
        keys = getkeys(keys)
        if keys == []:
            self.clear()
            return
        lastkey = keys.pop(-1)
        ans = self._data
        while keys:
            ans = ans[keys.pop(0)]
        del ans[lastkey]

    @setdoc.basic
    def __getitem__(self: Self, keys: tuple | int | str) -> Any:
        keys: list
        key: Any
        ans: Any
        keys = getkeys(keys)
        ans = self._data
        for key in keys:
            ans = ans[key]
        ans = getvalue(ans)
        return ans

    @setdoc.basic
    def __setitem__(self: Self, keys: tuple | int | str, value: Any) -> None:
        lastkey: Any
        keys: list
        data: Any
        target: Any
        k: Any
        keys = getkeys(keys)
        if keys == []:
            self.data = value
            return
        lastkey = keys.pop()
        data = getdict(self._data)
        target = data
        for k in keys:
            if isinstance(target, dict):
                target = target.setdefault(k, {})
            else:
                target = target[k]
        target[lastkey] = value
        self.data = data

    @property
    @setdoc.basic
    def data(self: Self) -> frozendict[str, Any]:
        return getdict(self._data, freeze=True)

    @data.setter
    def data(self: Self, value: Any) -> None:
        self._data = getdict(dict(value))

    @data.deleter
    def data(self: Self) -> None:
        self.clear()

    def dump(self: Self, stream: Any, **kwargs: Any) -> None:
        "This method dumps the data into a byte stream."
        tomli_w.dump(self._data, stream, **kwargs)

    def dumpintofile(self: Self, file: str, **kwargs: Any) -> None:
        "This method dumps the data into a file."
        with open(file, "wb") as stream:
            self.dump(stream, **kwargs)

    def dumps(self: Self, **kwargs: Any) -> str:
        "This method dumps the data as a string."
        return tomli_w.dumps(self._data, **kwargs)

    def get(self: Self, *keys: int | str, default: Any = None) -> Any:
        "This method returns self[*keys] if that exists, otherwise default."
        try:
            return self[keys]
        except KeyError:
            return default

    @classmethod
    def load(cls: type, stream: Any, **kwargs: Any) -> Self:
        "This classmethod loads data from byte stream."
        data: dict
        ans: Self
        data = tomllib.load(stream, **kwargs)
        ans = cls(data)
        return ans

    @classmethod
    def loadfromfile(cls: type, file: str, **kwargs: Any) -> Self:
        "This classmethod loads data from file."
        with open(file, "rb") as stream:
            return cls.load(stream, **kwargs)

    @classmethod
    def loads(cls: type, string: str, **kwargs: Any) -> Self:
        "This classmethod loads data from string."
        data: dict
        ans: Self
        data = tomllib.loads(string)
        ans = cls(data, **kwargs)
        return ans

    def setdefault(self: Self, *keys: int | str, default: Any) -> Any:
        "This method returns self[*keys] after setting it to default if previously absent."
        ans: Any
        try:
            ans = self[keys]
        except:
            self[keys] = default
            ans = self[keys]
        return ans


Holder: type = TOMLHolder
